import glob
import logging
import os
import os.path as osp
import shutil
import time

import seutils

import qondor

logger = logging.getLogger("qondor")


class CMSSW(object):
    """docstring for CMSSW"""

    default_local_rundir = "/tmp/qondor"
    default_remote_rundir = "."

    @staticmethod
    def extract_tarball(tarball, outdir=None):
        if outdir is None:
            outdir = (
                CMSSW.default_remote_rundir
                if qondor.BATCHMODE
                else CMSSW.default_local_rundir
            )
            logger.warning("Will extract %s to default: %s", tarball, outdir)
            try:
                qondor.utils.create_directory(outdir)
            except Exception:
                logger.warning("Failed to create %s; attempting here (.)", outdir)
                outdir = "."
                qondor.utils.create_directory(outdir)
        else:
            qondor.utils.create_directory(outdir)
        cmssw_dir = qondor.utils.extract_tarball_cmssw(tarball, outdir=outdir)
        cmssw_src = osp.abspath(osp.join(cmssw_dir, "src"))
        return cmssw_src

    @classmethod
    def from_tarball(cls, tarball, scram_arch=None, outdir=None):
        if seutils.has_protocol(tarball):
            logger.info("Tarball %s seems to be located remote; copying", tarball)
            dst = osp.abspath(osp.basename(tarball))
            if osp.isfile(dst):
                logger.warning("Using pre-existing tarball %s", dst)
            else:
                seutils.cp(tarball, dst)
            tarball = dst
        cmssw_src = cls.extract_tarball(tarball, outdir)
        # See if the tarball was already compiled with some scram_arch, if so use it
        if scram_arch is None:
            compiled_arches = glob.glob(osp.join(cmssw_src, "../bin/slc*"))
            if compiled_arches:
                scram_arch = osp.basename(compiled_arches[0])
                logger.warning(
                    "Detected tarball was compiled with arch %s, using it", scram_arch
                )
        return cls(cmssw_src, scram_arch)

    @classmethod
    def for_release(cls, release, scram_arch=None, outdir=".", renew=True):
        logger.info("Setting up CMSSW %s, arch %s in %s", release, scram_arch, outdir)
        # Make sure there is a directory in which to create the new CMSSW
        qondor.utils.create_directory(outdir)
        with qondor.utils.switchdir(outdir):
            # Check if there is an existing CMSSW
            existing = glob.glob("*{0}*".format(release))
            if existing:
                existing = existing[0]
                if renew:
                    logger.warning(
                        "Detected existing CMSSW %s; deleting it", osp.abspath(existing)
                    )
                    shutil.rmtree(existing)
                else:
                    logger.warning(
                        "Detected existing CMSSW %s; returning existing instead",
                        osp.abspath(existing),
                    )
                    return cls(osp.join(existing, "src"), scram_arch)
            # Make fresh release
            qondor.utils.run_multiple_commands(
                [
                    "shopt -s expand_aliases",
                    "source /cvmfs/cms.cern.ch/cmsset_default.sh",
                    "cmsrel {0}".format(release),
                ],
                env=qondor.utils.get_clean_env(),
            )
            # Get the src of the newly setup CMSSW
            cmssw_src = osp.abspath(glob.glob("*{0}*/src".format(release))[0])
        cmssw = cls(cmssw_src, scram_arch)
        # A freshly setup CMSSW won't need to be renamed
        cmssw._is_renamed = True
        return cmssw

    def __init__(self, cmssw_path, scram_arch=None):
        super(CMSSW, self).__init__()
        cmssw_path = cmssw_path.rstrip("/")
        if cmssw_path.endswith("src"):
            cmssw_path = osp.dirname(cmssw_path)
        self.cmssw_path = osp.abspath(cmssw_path)
        self.cmssw_src = osp.join(self.cmssw_path, "src")
        if scram_arch is None:
            compiled_arches = glob.glob(osp.join(self.cmssw_src, "../bin/slc*"))
            if compiled_arches:
                self.scram_arch = osp.basename(compiled_arches[0])
                logger.warning(
                    "Detected CMSSW was compiled with arch %s, using it",
                    self.scram_arch,
                )
            else:
                try:
                    logger.warning("Attempting to find scram_arch...")
                    print(osp.basename(osp.normpath(osp.join(self.cmssw_src, ".."))))
                    self.scram_arch = qondor.get_arch(
                        osp.basename(osp.normpath(osp.join(self.cmssw_src, "..")))
                    )
                except RuntimeError:
                    logger.warning(
                        "Taking SCRAM_ARCH from environment; may mismatch with"
                        " CMSSW version of %s",
                        self.cmssw_src,
                    )
                    self.scram_arch = os.environ["SCRAM_ARCH"]
        else:
            self.scram_arch = scram_arch
        self._is_renamed = False
        self._is_externallinks = False

    def rename_project(self):
        if self._is_renamed:
            return
        self._is_renamed = True
        logger.info("Renaming project %s", self.cmssw_src)
        self.run_commands_nocmsenv(["scram b ProjectRename"])

    def scramb_externallinks(self):
        if self._is_externallinks:
            return
        self._is_externallinks = True
        logger.info("Doing scram b ExternalLinks %s", self.cmssw_src)
        self.run_commands_nocmsenv(["scram b ExternalLinks"])

    def cmsrun(self, cfg_file, **kwargs):
        """
        Specifically runs a cmsRun command. Expects a python file, and other
        keyword arguments are parsed as "keyword=value".
        The key "inputFiles" is treated differently: A list is expected,
        and every item is added as "inputFiles=item".
        The key "outputFile" is also treated differently: The passed string
        is formatted with keys "proc_id", "cluster_id", and "datestr" for
        convenience.
        """
        cmd = ["cmsRun", cfg_file]
        if "inputFiles" in kwargs:
            inputFiles = kwargs.pop("inputFiles")
            if qondor.utils.is_string(inputFiles):
                inputFiles = [inputFiles]
            for inputFile in inputFiles:
                cmd.append("inputFiles=" + inputFile)
        if "outputFile" in kwargs:
            outputFile = kwargs.pop("outputFile")
            cmd.append(
                "outputFile="
                + outputFile.format(
                    proc_id=qondor.get_proc_id(),
                    cluster_id=qondor.get_cluster_id(),
                    datestr=time.strftime("%b%d"),
                )
            )
        cmd.extend(["{0}={1}".format(k, v) for k, v in kwargs.items()])
        self.run_command(cmd)

    def run_command(self, cmd):
        """
        Mostly legacy; use run_commands instead
        """
        return self.run_commands([cmd])

    def run_commands(self, cmds):
        """
        Main/Public method: Much like run_commands_nocmsenv, but includes cmsenv
        This is intended to be called with cmsRun/cmsDriver.py commands
        """
        return self.run_commands_nocmsenv(["cmsenv"] + cmds)

    def run_commands_nocmsenv(self, cmds):
        """
        Preprends the basic CMSSW environment setup, and executes a set of
        commands in a clean environment
        """
        self.rename_project()
        with qondor.utils.switchdir(self.cmssw_src):
            return qondor.utils.run_multiple_commands(
                [
                    "shopt -s expand_aliases",
                    "source /cvmfs/cms.cern.ch/cmsset_default.sh",
                    "export SCRAM_ARCH={0}".format(self.scram_arch),
                ]
                + cmds,
                env=qondor.utils.get_clean_env(),
            )

    def make_tarball(
        self,
        outdir=".",
        tag=None,
        renew=True,
        exclude=None,
        exclude_vcs=True,
        exclude_caches_all=True,
        include=None,
    ):
        """
        Makes a tarball out of the CMSSW distribution of this class
        """
        cmssw_path = osp.abspath(osp.join(self.cmssw_src, ".."))
        # Determine location of the output tarball
        dst = osp.join(
            osp.abspath(outdir),
            (
                osp.basename(cmssw_path).strip("/")
                + ("" if tag is None else "_" + tag)
                + ".tar.gz"
            ),
        )
        if osp.isfile(dst):
            if renew:
                logger.warning("Removing existsing %s", dst)
                os.remove(dst)
            else:
                raise OSError("{0} already exists".format(dst))
        logger.warning("Tarballing {0} ==> {1}".format(cmssw_path, dst))
        if exclude is None:
            exclude = []
        exclude.append("tmp")
        with qondor.utils.switchdir(osp.dirname(cmssw_path)):
            cmd = [
                "tar",
                "-zcvf",
                dst,
            ]
            if include:
                if qondor.utils.is_string(include):
                    include = [include]
                cmd.extend(include)
            if exclude_vcs:
                cmd.append(
                    "--exclude-vcs",
                )
            if exclude_caches_all:
                cmd.append(
                    "--exclude-caches-all",
                )
            cmd.extend(["--exclude={0}".format(e) for e in exclude])
            cmd.append(osp.basename(cmssw_path))
            qondor.utils.run_command(cmd)
        return dst

    def pip(self, *args):
        """
        Installs a pypi package in this CMSSW environment using the scram-pip script.
        Any arguments are used for the pip command
        """
        scrampip = osp.join(qondor.INCLUDE_DIR, "scram-pip")
        cmd = [scrampip] + list(args)
        self.run_commands([cmd, "cmsenv"])

    def make_chunk_rootfile(self, src, first, last, tree="auto", dst=None):
        """
        Round-about way of calling seutils.make_chunk_rootfile in this CMSSW environment.
        First installs seutils in this CMSSW using scram pip, then calls seu-takechunkroot
        to perform the actual splitting.
        This is a rather hacky solution.
        """
        self.pip("seutils", '-p="--ignore-installed"')
        cmd = "seu-takechunkroot {src} -f {first} -l {last} -t {tree}".format(
            src=src, first=first, last=last, tree=tree
        )
        if dst:
            cmd += " -d " + dst
        self.run_command(cmd)

    def get_chunk(self, chunk, dst, tree="auto"):
        """
        Roundabout way of calling seutils.hadd_chunk_entries in this CMSSW environment.
        """
        return seutils.hadd_chunk_entries(
            chunk, dst, tree=tree, file_split_fn=self.make_chunk_rootfile
        )



# -*- coding: utf-8 -*-
import glob
import logging
import os.path as osp
import re

import qondor

logger = logging.getLogger("qondor")


def build_resubmission(jsonfile):
    if osp.isdir(jsonfile):
        # Treat it as a rundir
        jsonfile = osp.join(jsonfile, "submission_latest.json")
    if not osp.isfile(jsonfile):
        raise Exception("Pass a path to a rundir or a specific json file")
    logger.debug("Loading %s", jsonfile)
    with open(jsonfile, "r") as f:
        submission = qondor._json_load_byteified(f)
    resub = Resubmission(jsonfile, **submission)
    return resub


class ClusterFlags:
    PARTIALLY_NOT_STARTED = 1
    COMPLETED_UNSUCCESSFULLY = 10
    COMPLETED_SUCCESSFULLY = 11

    @classmethod
    def strrep(cls, i):
        for key in cls.__dict__:
            if cls.__dict__[key] == i:
                return key
        else:
            raise Exception("No such state: {}".format(i))


class JobFlags:
    GOT_SUBMITTED = 0
    NOT_STARTED = 1
    IN_HTCONDOR = 2
    HTCONDOR_ERROR = 3
    EXITED = 4
    GOOD_EXITCODE = 5
    BAD_EXITCODE = 6
    NO_EXITCODE = 7


SUCCESSES = [JobFlags.GOOD_EXITCODE]
FAILURES = [JobFlags.HTCONDOR_ERROR, JobFlags.BAD_EXITCODE, JobFlags.NO_EXITCODE]


def int_to_str(status, color=True):
    for key in JobFlags.__dict__:
        if JobFlags.__dict__[key] == status:
            if color:
                if status in SUCCESSES:
                    key = qondor.colored(key, "green")
                elif status in FAILURES:
                    key = qondor.colored(key, "red")
                elif status in [JobFlags.NOT_STARTED, JobFlags.IN_HTCONDOR]:
                    key = qondor.colored(key, "yellow")
            return key
    else:
        raise Exception("No such flag: {}".format(status))


def ints_to_str(statuss, color=True):
    return ",".join([int_to_str(s, color) for s in statuss])


class Job(object):
    """
    Simple container to store the current state of a job
    """

    def __init__(self, cluster_id, proc_id, flags=None):
        self.cluster_id = cluster_id
        self.proc_id = proc_id
        self.flags = [] if flags is None else flags
        self.htcondor_status = None
        self.htcondor_status_str = None

    def is_running(self):
        return JobFlags.IN_HTCONDOR in self.flags and self.htcondor_status not in [5, 6]

    def is_failed(self):
        if self.is_running():
            return False
        for failure in FAILURES:
            if failure in self.flags:
                return True
        return False

    def is_success(self):
        return JobFlags.GOOD_EXITCODE in self.flags

    def __repr__(self):
        r = "{}.{} {}".format(self.cluster_id, self.proc_id, ints_to_str(self.flags))
        if self.htcondor_status_str is not None:
            r += " " + self.htcondor_status_str
        return r


class Resubmission(object):
    def __init__(self, jsonfile, **kwargs):
        self.jsonfile = osp.abspath(jsonfile)
        self.rundir = osp.dirname(self.jsonfile)
        self.__dict__.update(**kwargs)
        self.out_files = glob.glob(self.rundir + "/out_*.txt")
        self.err_files = glob.glob(self.rundir + "/err_*.txt")
        self.log_files = glob.glob(self.rundir + "/log_*.txt")
        self.exitcode_files = glob.glob(self.rundir + "/exitcode_*.txt")
        self._jobs_cache = []
        self._jobs_read = False

    def iter_only_failed_jobs(self):
        for sub, cluster_id, njobs, proc_ids, jobs in self._jobs_cache:
            failed_jobs = list(filter(lambda j: j.is_failed(), jobs))
            proc_ids = [j.proc_id for j in failed_jobs]
            yield sub, cluster_id, njobs, proc_ids, failed_jobs

    def jobs(self, only_failed=False):
        if not self._jobs_read:
            self.read_jobs()
        if only_failed:
            return list(self.iter_only_failed_jobs())
        return self._jobs_cache

    def job_objects(self, only_failed=False):
        for _, _, _, _, jobs in self.jobs(only_failed):
            yield jobs

    def read_jobs(self):
        self._jobs_read = True
        self._jobs_cache = []
        all_queued_jobs = qondor.schedd.get_jobs()
        for sub, cluster_id, njobs, proc_ids in self.submitted:
            name = sub["environment"]["QONDORCLUSTERNAME"]
            logger.debug(
                "Found job %s: cluster_id=%s, njobs=%s, proc_ids=%s",
                name,
                cluster_id,
                njobs,
                proc_ids,
            )
            # Gather the files for this submission created by htcondor
            out_files = []
            err_files = []
            log_files = []
            exitcode_files = []
            for filetype in ["out", "err", "log", "exitcode"]:
                thelist = locals()[filetype + "_files"]
                pat = re.compile(r"{}_{}_\d+_\d+\.txt".format(filetype, name))
                for file in getattr(self, filetype + "_files"):
                    if pat.match(file):
                        thelist.append(filetype)
            # Get the queued jobs for this submission in htcondor
            queued_jobs = list(
                filter(lambda job: job.cluster_id == cluster_id, all_queued_jobs)
            )

            def find_in_queue(job):
                for queued_job in queued_jobs:
                    if queued_job.proc_id == job.proc_id:
                        return queued_job
                else:
                    return None

            # Per job in the submission, determine the flags
            jobs = []
            for proc_id in proc_ids:
                job = Job(cluster_id, proc_id)
                # Check if the job exists in the htcondor queue
                queued_job = find_in_queue(job)
                if not (queued_job is None):
                    job.flags.append(JobFlags.IN_HTCONDOR)
                    job.htcondor_status = queued_job.status
                    job.htcondor_status_str = queued_job.status_str()
                    # htcondor status's 5 and 6 mean an error
                    if job.htcondor_status == 5 or job.htcondor_status == 6:
                        job.flags.append(JobFlags.HTCONDOR_ERROR)
                # Check if a .log file was created for the job
                if osp.isfile(
                    osp.join(
                        self.rundir,
                        "log_{}_{}_{}.txt".format(name, cluster_id, proc_id),
                    )
                ):
                    job.flags.append(JobFlags.GOT_SUBMITTED)
                else:
                    job.flags.append(JobFlags.NOT_STARTED)
                # Check if a .out file was created for the job; if so, the job must have exited
                if osp.isfile(
                    osp.join(
                        self.rundir,
                        "out_{}_{}_{}.txt".format(name, cluster_id, proc_id),
                    )
                ):
                    job.flags.append(JobFlags.EXITED)
                # Check if an exitcode file was created for the job, and if so check wether
                # the exitcode equals 0
                exitcode_file = osp.join(
                    self.rundir,
                    "exitcode_{}_{}_{}.txt".format(name, cluster_id, proc_id),
                )
                if osp.isfile(exitcode_file):
                    with open(exitcode_file, "r") as f:
                        exitcode = int(f.read().strip())
                    if exitcode == 0:
                        job.flags.append(JobFlags.GOOD_EXITCODE)
                    else:
                        job.flags.append(JobFlags.BAD_EXITCODE)
                else:
                    job.flags.append(JobFlags.NO_EXITCODE)
                jobs.append(job)
            sub_tuple = (sub, cluster_id, njobs, proc_ids, jobs)
            self._jobs_cache.append(sub_tuple)

    def print_jobs(self, only_failed=False, summary=False):
        r = []
        n_good = 0
        n_failed = 0
        n_running = 0
        n_total = 0
        for sub, cluster_id, njobs, proc_ids, jobs in self.jobs(only_failed):
            if not len(jobs):
                continue
            name = sub["environment"]["QONDORCLUSTERNAME"]
            i_cluster = int(sub["environment"]["QONDORICLUSTER"])
            r.append("{: <3} {}".format(i_cluster, name))
            for proc_id, job in zip(proc_ids, jobs):
                r.append("  " + repr(job))
                if job.is_success():
                    n_good += 1
                elif job.is_failed():
                    n_failed += 1
                else:
                    n_running += 1
                n_total += 1
        if summary:
            r.append(
                "total: {}, good: {} ({:.1f}%), failed: {} ({:.1f}%), running: {} ({:.1f}%)".format(
                    n_total,
                    n_good,
                    (100.0 * n_good) / n_total,
                    n_failed,
                    (100.0 * n_failed) / n_total,
                    n_running,
                    (100.0 * n_running) / n_total,
                )
            )
        print("\n".join(r))

    def resubmit(self, cli=True):
        with qondor.utils.switchdir(self.rundir):
            session = qondor.submit.Session("resubmission")
            for sub_orig, cluster_id, njobs, proc_ids, jobs in self.jobs(
                only_failed=True
            ):
                sub = sub_orig.copy()
                for job in jobs:
                    logger.info("Submitting job %s", job)
                    sub["environment"]["QONDOR_PROC_ID_RESUBMISSION"] = job.proc_id
                    session.submittables.append((sub, 1))
            session.submit(cli)



# -*- coding: utf-8 -*-
import json
import logging
import os
import os.path as osp
import pprint
import re
import shutil
from datetime import datetime

import seutils

import qondor

logger = logging.getLogger("qondor")


# _____________________________________________________________________
# Interface to translate python code (typically in a file) into job submissions


def split_runcode_submitcode(lines):
    """
    Splits a list of lines into two strings, the python to run in the job
    and the other the python code to submit the jobs.
    """
    runcode_lines = []
    submitcode_lines = []
    is_runcode_mode = True
    is_submitcode_mode = False
    for line in lines:
        line = line.strip("\n")
        if line.startswith('"""# submit'):
            if is_submitcode_mode:
                raise Exception(
                    "Encountered submit code opening tag, but was already in submit mode"
                )
            # Toggle the mode
            is_runcode_mode = not (is_runcode_mode)
            is_submitcode_mode = not (is_submitcode_mode)
            continue
        elif line.startswith('"""# endsubmit'):
            if is_runcode_mode:
                raise Exception(
                    "Encountered submit code closing tag, but was already not in submit mode"
                )
            # Toggle the mode
            is_runcode_mode = not (is_runcode_mode)
            is_submitcode_mode = not (is_submitcode_mode)
            continue
        else:
            if is_runcode_mode:
                runcode_lines.append(line)
            else:
                submitcode_lines.append(line)
    submitcode = "\n".join(submitcode_lines)
    runcode = "\n".join(runcode_lines)
    return runcode, submitcode


def split_runcode_submitcode_file(filename):
    """
    Wrapper for split_runcode_submitcode that opens up the file first
    """
    with open(filename, "r") as f:
        return split_runcode_submitcode(f.readlines())


def exec_wrapper(code, scope):
    """
    Python 2 has problems with function definitions in the same scope as the exec.
    Solution: Run exec in a subscope with this wrapper function.
    """
    exec(code, scope)


class StopProcessing(Exception):
    """
    Special exception to stop execution inside an exec statement
    """

    pass


def submit_python_job_file(
    filename, cli=False, njobsmax=None, run_args=None, return_first_cluster=False
):
    """
    Builds submission clusters from a python job file
    """
    runcode, submitcode = split_runcode_submitcode_file(filename)
    # Run the submitcode
    # First create exec scope dict, with a few handy functions
    _first_cluster_ptr = [None]
    n_calls_to_submit_fn = [0]
    session = Session(name=osp.basename(filename).replace(".py", ""))
    # Place to store 'global' pip installs
    pips = []

    def htcondor_setting(key, value):
        session.htcondor(key, value)

    def pip_fn(package, install_instruction=None):
        if install_instruction:
            pips.append((package, install_instruction))
        else:
            pips.append(package)

    def submit_fn(*args, **kwargs):
        kwargs.setdefault("session", session)
        kwargs.setdefault("run_args", run_args)
        kwargs["pips"] = kwargs.get("pips", []) + pips  # Add 'global' installs
        njobs = kwargs.get("njobs", 1)
        cluster = Cluster(runcode, *args, **kwargs)
        if return_first_cluster:
            _first_cluster_ptr[0] = cluster
            raise StopProcessing
        session.add_submission(cluster, cli=cli, njobsmax=njobsmax, njobs=njobs)
        n_calls_to_submit_fn[0] += 1

    def submit_now_fn():
        session.submit(cli, njobsmax=njobsmax)

    exec_scope = {
        "qondor": qondor,
        "session": session,
        "htcondor": htcondor_setting,
        "submit": submit_fn,
        "pip": pip_fn,
        "submit_now": submit_now_fn,
        "runcode": runcode,
        "run_args": run_args,
        "cli": cli,
        "njobsmax": njobsmax,
        "return_first_cluster": return_first_cluster,
    }
    logger.info("Running submission code now")
    if return_first_cluster:
        try:
            exec_wrapper(submitcode, exec_scope)
        except StopProcessing:
            pass
        cluster = _first_cluster_ptr[0]
        if cluster is None:
            raise Exception("No cluster was submitted in the submit code")
        return cluster
    else:
        exec_wrapper(submitcode, exec_scope)
        # Special case: There was no call to submit
        # Just submit 1 job in that case
        if n_calls_to_submit_fn[0] == 0:
            logger.info(
                "No calls to submit() were made in the submit code; submitting 1 job"
            )
            cluster = Cluster(runcode, session=session, run_args=run_args, njobs=1)
            session.add_submission(cluster, cli=cli, njobsmax=njobsmax, njobs=1)
        # Do a submit call in case it wasn't done yet (if submit_now was called in the submit code this is a no-op)
        session.submit(cli, njobsmax=njobsmax)


def get_first_cluster(filename):
    """
    Returns the first cluster that would be submitted if the python job file
    would be submitted
    """
    return submit_python_job_file(filename, return_first_cluster=True)


# _____________________________________________________________________
# Submission interface:
# Deals with building submission dicts and submitting them to htcondor

# Run environments: Currently just two options (sl6 and sl7)
RUN_ENVS = {
    "sl7-py27": [
        'export pipdir="/cvmfs/sft.cern.ch/lcg/releases/pip/19.0.3-06476/x86_64-centos7-gcc7-opt"',
        'export SCRAM_ARCH="slc7_amd64_gcc820"',
        "source /cvmfs/sft.cern.ch/lcg/contrib/gcc/7/x86_64-centos7/setup.sh",
        "source /cvmfs/sft.cern.ch/lcg/releases/LCG_95/ROOT/6.16.00/x86_64-centos7-gcc7-opt/ROOT-env.sh",
        'export PATH="${pipdir}/bin:${PATH}"',
        'export PYTHONPATH="${pipdir}/lib/python2.7/site-packages/:${PYTHONPATH}"',
        'pip(){ ${pipdir}/bin/pip "$@"; }  # To avoid any local pip installations',
    ],
    "sl6-py27": [
        'export pipdir="/cvmfs/sft.cern.ch/lcg/releases/pip/19.0.3-06476/x86_64-slc6-gcc7-opt"',
        'export SCRAM_ARCH="slc6_amd64_gcc700"',
        "source /cvmfs/sft.cern.ch/lcg/contrib/gcc/7/x86_64-slc6-gcc7-opt/setup.sh",
        "source /cvmfs/sft.cern.ch/lcg/releases/LCG_95/ROOT/6.16.00/x86_64-slc6-gcc7-opt/ROOT-env.sh",
        'export PATH="${pipdir}/bin:${PATH}"',
        'export PYTHONPATH="${pipdir}/lib/python2.7/site-packages/:${PYTHONPATH}"',
        'pip(){ ${pipdir}/bin/pip "$@"; }  # To avoid any local pip installations',
    ],
    "centos7-py36": [
        "source /cvmfs/sft.cern.ch/lcg/views/LCG_96python3/x86_64-centos7-gcc8-opt/setup.sh",
    ],
}


def get_default_sub(submission_time=None):
    """
    Returns the default submission dict (the equivalent of a .jdl file)
    to be used by the submitter.
    """
    if submission_time is None:
        submission_time = datetime.now()
    sub = {
        "universe": "vanilla",
        "output": "out_$(Cluster)_$(Process).txt",
        "error": "err_$(Cluster)_$(Process).txt",
        "log": "log_$(Cluster)_$(Process).txt",
        "should_transfer_files": "YES",
        "environment": {
            "QONDOR_BATCHMODE": "1",
            "CONDOR_CLUSTER_NUMBER": "$(Cluster)",
            "CONDOR_PROCESS_ID": "$(Process)",
            "CLUSTER_SUBMISSION_TIMESTAMP": submission_time.strftime(
                qondor.TIMESTAMP_FMT
            ),
        },
    }
    # Try to set some more things
    try:
        sub["x509userproxy"] = os.environ["X509_USER_PROXY"]
    except KeyError:
        try:
            sub["x509userproxy"] = qondor.utils.run_command(
                ["voms-proxy-info", "-path"]
            )[0].strip()
            logger.info(
                'Set x509userproxy to "%s" based on output from voms-proxy-info',
                sub["x509userproxy"],
            )
        except Exception:
            logger.warning(
                "Could not find a x509userproxy to pass; manually "
                "set the htcondor variable 'x509userproxy' if your "
                "htcondor setup requires it."
            )
    try:
        sub["environment"]["USER"] = os.environ["USER"]
    except KeyError:
        # No user specified, no big deal
        pass
    return sub


def update_sub(sub, other):
    """
    Merges a submission dict, attempting not to overwrite some keys but rather append them
    """
    # First copy all
    r = dict(sub, **other)
    # Merge some things more carefully:
    # Environment
    r["environment"] = dict(sub.get("environment", {}), **other.get("environment", {}))
    # Input files
    files = []
    if "transfer_input_files" in sub:
        files += sub["transfer_input_files"].split(",")
    if "transfer_input_files" in other:
        files += other["transfer_input_files"].split(",")
    if files:
        r["transfer_input_files"] = ",".join(files)
    return r


class Session(object):
    """
    Over-arching object that controls submission of a number of clusters
    """

    def __init__(self, name=None):
        self.submission_time = datetime.now()
        name = "qondor_" + name if name else "qondor"
        self.rundir = osp.abspath(
            "{}_{}".format(name, self.submission_time.strftime(qondor.TIMESTAMP_FMT))
        )
        self.transfer_files = []
        self._created_python_module_tarballs = {}
        self._i_seutils_tarball = 0
        self.submittables = []
        self._njobs_submitted = 0
        self.htcondor_settings = get_default_sub(self.submission_time)
        self.htcondor_settings["+QondorRundir"] = '"' + self.rundir + '"'
        self._fixed_cmsconnect_specific_settings = False
        # Storage for submitted jobs
        self.submitted = []

    def htcondor(self, key, value):
        """
        Shortcut to add htcondor settings
        """
        self.htcondor_settings[key] = value

    def fix_cmsconnect_specific_settings_once(self, cli):
        """
        Potentially process cmsconnect specific settings
        Changes self.htcondor_settings once
        """
        if self._fixed_cmsconnect_specific_settings:
            return
        self._fixed_cmsconnect_specific_settings = True
        blacklist = self.htcondor_settings.pop("cmsconnect_blacklist", None)
        whitelist = self.htcondor_settings.pop("cmsconnect_whitelist", None)
        if os.uname()[1] == "login.uscms.org" or os.uname()[1] == "login-el7.uscms.org":
            qondor.logger.warning("Detected CMS Connect; loading specific settings")
            cmsconnect_settings(
                self.htcondor_settings,
                cli=cli,
                blacklist=blacklist,
                whitelist=whitelist,
            )

    def dump_seutils_cache(self):
        """
        Dumps the seutils cache to a tarball, to be included in the job.
        """
        proposed_tarball = osp.join(
            self.rundir, "seutils-cache-{}.tar.gz".format(self._i_seutils_tarball)
        )
        # If no new seutils calls were made, the tarball won't be remade
        with seutils.drymode_context(qondor.DRYMODE):
            actual_tarball = seutils.tarball_cache(
                proposed_tarball, only_if_updated=True
            )
        if actual_tarball == proposed_tarball:
            self._i_seutils_tarball += 1
        logger.info("Using seutils-cache tarball %s", actual_tarball)
        return actual_tarball

    def handle_python_package_tarballs(self, cluster):
        # Put in python package tarballs required for the code in the job
        for package, install_instruction in cluster.pips:
            # Packages with a specific version should always be installed from pypi
            for c in ["<", "=", ">"]:
                if c in package:
                    install_instruction = "pypi"
            # Determine whether package was installed editably
            if install_instruction == "auto" and qondor.utils.dist_is_editable(package):
                install_instruction = "editable"
            # If package was installed editably, tarball it up and include it
            if install_instruction == "editable":
                # Create the tarball if it wasn't already created
                if package not in self._created_python_module_tarballs:
                    self._created_python_module_tarballs[
                        package
                    ] = qondor.utils.tarball_python_module(package, outdir=self.rundir)
                # Add the tarball as input file for this cluster
                cluster.transfer_files[
                    "_packagetarball_{}".format(package)
                ] = self._created_python_module_tarballs[package]

    def add_submission(self, cluster, cli=True, njobs=1, njobsmax=None):
        if njobsmax:
            njobs = min(njobs, njobsmax - self._njobs_submitted)
        if njobsmax and njobs == 0:
            logger.debug(
                "Not adding submission for cluster %s - reached njobsmax %s",
                cluster.i_cluster,
                njobsmax,
            )
            return
        self._njobs_submitted += njobs
        qondor.utils.create_directory(self.rundir)
        cluster.rundir = self.rundir
        # Possibly create tarballs out of required python packages
        self.handle_python_package_tarballs(cluster)
        # Possibly create tarball for the seutils cache and include it in the job
        if seutils.USE_CACHE:
            cluster.transfer_files["seutils-cache"] = self.dump_seutils_cache()
        # Dump cluster contents to file and build scope
        cluster.runcode_to_file()
        cluster.sh_entrypoint_to_file()
        cluster.scope_to_file()
        # Compile the submission dict
        # Base off of global settings only for python-binding mode
        # For the condor_submit cli method, it's better to just write the global keys once at the top of the file
        # Little unsafe, if you modiy these 'global' settings in one job they propegate to all the next ones,
        # but otherwise submission of many jobs becomes very slow
        sub = {
            "output": "out_{}_$(Cluster)_$(Process).txt".format(cluster.name),
            "error": "err_{}_$(Cluster)_$(Process).txt".format(cluster.name),
            "log": "log_{}_$(Cluster)_$(Process).txt".format(cluster.name),
        }
        sub["executable"] = osp.abspath(cluster.sh_entrypoint_filename)
        sub["environment"] = {}
        sub["environment"]["QONDORICLUSTER"] = str(cluster.i_cluster)
        sub["environment"]["QONDORCLUSTERNAME"] = str(cluster.name)
        sub["environment"].update(cluster.env)
        # Overwrite htcondor keys defined in the preprocessing
        sub.update(cluster.htcondor)
        # Flatten files into a string, excluding files on storage elements
        transfer_files = self.transfer_files + [
            f for f in cluster.transfer_files.values() if not seutils.has_protocol(f)
        ]
        if len(transfer_files) > 0:
            sub["transfer_input_files"] = ",".join(transfer_files)
        sub = update_sub(sub, cluster.htcondor)
        # Plugin the global and cmsconnect settings in now
        self.fix_cmsconnect_specific_settings_once(cli)
        sub = update_sub(self.htcondor_settings, sub)
        logger.info(
            "Prepared submission dict for cluster %s:\n%s",
            cluster.i_cluster,
            pprint.pformat(sub),
        )
        # Add it to the submittables
        self.submittables.append((sub, njobs))

    def submit_pythonbindings(self, njobsmax=None):
        qondor.utils.check_proxy()
        if not self.submittables:
            return
        import htcondor

        if njobsmax is None:
            njobsmax = 1e7
        n_jobs_summed = sum([njobs for _, njobs in self.submittables])
        n_jobs_total = min(n_jobs_summed, njobsmax)
        logger.info("Submitting all jobs; %s out of %s", n_jobs_total, n_jobs_summed)
        schedd = qondor.schedd.get_best_schedd()
        n_jobs_todo = n_jobs_total
        ads = []
        with qondor.utils.switchdir(self.rundir):
            with qondor.schedd._transaction(schedd) as transaction:
                submit_object = htcondor.Submit()
                for sub_orig, njobs in self.submittables:
                    sub = (
                        sub_orig.copy()
                    )  # Keep original dict intact? Global settings already contained
                    sub["environment"] = qondor.schedd.format_env_htcondor(
                        sub["environment"]
                    )
                    njobs = min(njobs, n_jobs_todo)
                    n_jobs_todo -= njobs
                    # Load the dict into the submit object
                    for key in sub:
                        submit_object[key] = sub[key]
                    new_ads = []
                    cluster_id = (
                        int(submit_object.queue(transaction, njobs, new_ads))
                        if not qondor.DRYMODE
                        else 0
                    )
                    logger.warning(
                        "Submitted %s jobs for i_cluster %s (%s) to htcondor cluster %s",
                        len(new_ads) if not qondor.DRYMODE else njobs,
                        sub_orig["environment"]["QONDORICLUSTER"],
                        sub_orig["environment"]["QONDORCLUSTERNAME"],
                        cluster_id,
                    )
                    ads.extend(new_ads)
                    self.submitted.append(
                        (
                            sub_orig,
                            cluster_id,
                            len(new_ads),
                            [ad["ProcId"] for ad in new_ads],
                        )
                    )
                    if n_jobs_todo == 0:
                        break
        logger.info(
            "Summary: Submitted %s jobs to cluster %s", n_jobs_total, cluster_id
        )

    def submit_cli(self, njobsmax=None):
        qondor.utils.check_proxy()
        if not self.submittables:
            return
        if njobsmax is None:
            njobsmax = 1e7
        n_jobs_summed = sum([njobs for _, njobs in self.submittables])
        n_jobs_total = min(n_jobs_summed, njobsmax)
        logger.info("Submitting all jobs; %s out of %s", n_jobs_total, n_jobs_summed)
        n_jobs_todo = n_jobs_total
        # Compile the .jdl file

        def get_cluster_nr(sub):
            return sub["environment"]["QONDORICLUSTER"]

        jdl_file = osp.join(
            self.rundir,
            "qondor_{}-{}.jdl".format(
                get_cluster_nr(self.submittables[0][0]),
                get_cluster_nr(self.submittables[-1][0]),
            ),
        )
        jdl_contents = []
        # Write the settings per job
        for sub, njobs in self.submittables:
            njobs = min(njobs, n_jobs_todo)
            n_jobs_todo -= njobs
            jdl_contents.append(
                "# Cluster {}".format(sub["environment"]["QONDORICLUSTER"])
            )
            # Dump the submission to a jdl file
            for key in sub.keys():
                val = sub[key]
                if key.lower() == "environment":
                    val = qondor.schedd.format_env_htcondor(val)
                jdl_contents.append("{} = {}".format(key, val))
            jdl_contents.append("queue {}\n".format(njobs))
            if n_jobs_todo == 0:
                break
        # Dump to file
        jdl_contents = "\n".join(jdl_contents)
        with qondor.utils.openfile(jdl_file, "w") as jdl:
            jdl.write(jdl_contents)
        logger.info("Compiled %s:\n%s", jdl_file, jdl_contents)
        # Run the actual submit command
        with qondor.utils.switchdir(self.rundir):
            output = qondor.utils.run_command(["condor_submit", osp.basename(jdl_file)])
        # Get some info from the condor_submit command, if not dry mode
        if qondor.DRYMODE:
            logger.warning("Summary: Submitted %s jobs to cluster 0", n_jobs_total)
        else:
            matches = re.findall(
                r"(\d+) job\(s\) submitted to cluster (\d+)", "\n".join(output)
            )
            if not len(matches):
                logger.error(
                    "condor_submit exited ok but could not determine where and how many jobs were submitted"
                )
            # Unfortunately we have to match the number of submitted jobs back to the submission dicts
            # with a while-loop and careful counting
            submittables = iter(self.submittables)
            for njobs_submitted, cluster_id in matches:
                njobs_submitted = int(njobs_submitted)
                logger.warning(
                    "Submitted %s jobs to cluster_id %s", njobs_submitted, cluster_id
                )
                njobs_assigned = 0
                while njobs_assigned < njobs_submitted:
                    sub, njobs_thissub = next(submittables)  # Get the next sub
                    proc_ids = list(
                        range(njobs_assigned, njobs_assigned + njobs_thissub)
                    )  # Build the list of proc_ids
                    njobs_assigned += njobs_thissub
                    self.submitted.append((sub, cluster_id, njobs_assigned, proc_ids))

    def submit(self, cli, *args, **kwargs):
        """
        Wrapper that just picks the specific submit method
        """
        if len(self.submittables) == 0:
            logger.warning("No jobs to be submitted")
            return
        qondor.utils.create_directory(self.rundir)
        # Run the submission code
        self.submit_cli(*args, **kwargs) if cli else self.submit_pythonbindings(
            *args, **kwargs
        )
        if not qondor.DRYMODE:
            # Dump some submission information to a .json file for possible resubmission later
            njobsmax = kwargs.get("njobsmax", None)
            submission_timestamp = qondor.get_submission_timestamp()
            submission = {
                "submission_timestamp": submission_timestamp,
                # 'submittables' : self.submittables,
                "njobsmax": njobsmax,
                "cli": cli,
                "submitted": self.submitted,
            }
            submission_jsonfile = osp.join(
                self.rundir, "submission_{}.json".format(submission_timestamp)
            )
            with open(submission_jsonfile, "w") as f:
                json.dump(submission, f)
            # Also copy this file to submission_latest.json for easier retrieval
            shutil.copyfile(
                submission_jsonfile, osp.join(self.rundir, "submission_latest.json")
            )
        # Clear the submittables
        self.submittables = []
        self.submitted = []


class Cluster(object):

    ICLUSTER = 0
    NAMES = set()

    def __init__(
        self,
        runcode,
        scope=None,
        env=None,
        pips=None,
        htcondor=None,
        run_args=None,
        run_env="sl7-py27",
        rundir=".",
        session=None,
        transfer_files=None,
        **kwargs
    ):
        self.i_cluster = self.__class__.ICLUSTER
        self.__class__.ICLUSTER += 1
        self.rundir = rundir
        self.runcode = runcode
        self.env = {} if env is None else env
        self.scope = {} if scope is None else scope
        self.htcondor = {} if htcondor is None else htcondor
        self.run_args = run_args
        if qondor.utils.is_string(run_env):
            run_env = RUN_ENVS[run_env]
        self.run_env = run_env
        self.transfer_files = {}
        if transfer_files:
            for f in transfer_files:
                self._add_transfer_file_str_or_tuple(f)
        self.session = session
        # Try to use a more human-readable name if it's given
        if "name" not in kwargs:
            self.name = "cluster{}".format(self.i_cluster)
        else:
            # If the name was already used before, append a number to it until it's unique
            name = kwargs["name"]
            if name in self.NAMES:
                name += "{}"
                i_attempt = 0
                while name.format(i_attempt) in self.NAMES:
                    i_attempt += 1
                name = name.format(i_attempt)
            self.__class__.NAMES.add(name)
            self.name = name
        logger.info("Using name %s", self.name)
        # Base filenames needed for the job
        self.runcode_filename = "{}.py".format(self.name)
        self.sh_entrypoint_filename = "{}.sh".format(self.name)
        self.scope_filename = "{}.json".format(self.name)
        # Process pip packages
        self.pips = []
        pips = [] if pips is None else pips

        def check_if_in_pips(package):
            """Checks if a package is already in the list"""
            package, _ = qondor.utils.pip_split_version(package)
            for pip in pips:
                if not qondor.utils.is_string(pip):
                    pip = pip[0]
                pip, _ = qondor.utils.pip_split_version(pip)
                if pip == package:
                    return True
            return False

        if not check_if_in_pips("qondor"):
            pips.append(("qondor", "auto"))
        if not check_if_in_pips("seutils"):
            pips.append(("seutils", "auto"))
        for pip in pips:
            if qondor.utils.is_string(pip):
                self.pips.append((pip, "auto"))
            else:
                self.pips.append((pip[0], pip[1]))
        # Add addtional keywords to the scope
        self.scope.update(kwargs)

    def _add_transfer_file_str_or_tuple(self, t):
        if qondor.utils.is_string(t):
            self.add_transfer_file(t)
        else:
            key, value = t
            self.add_transfer_file(value, key=key)

    def add_transfer_file(self, filename, key=None):
        """
        Adds a file to the self.transfer_files dict in order for it to be transferred
        """
        filename = osp.abspath(osp.expanduser(filename))
        if key is None:
            key = osp.basename(filename)
            # Make a unique key
            if key in self.transfer_files:
                key += "_{0}"
                for i in range(100):
                    if not key.format(i) in self.transfer_files:
                        key = key.format(i)
                        break
                else:
                    raise Exception("Could not make a unique key")
        self.transfer_files[key] = filename

    def runcode_to_file(self):
        self.runcode_filename = osp.join(self.rundir, self.runcode_filename)
        if osp.isfile(self.runcode_filename):
            raise OSError("{} exists".format(self.runcode_filename))
        self.transfer_files["runcode"] = self.runcode_filename
        logger.info(
            "Dumping python code for cluster %s to %s",
            self.i_cluster,
            self.runcode_filename,
        )
        if not (qondor.DRYMODE):
            with open(self.runcode_filename, "w") as f:
                f.write(self.runcode)

    def sh_entrypoint_to_file(self):
        self.sh_entrypoint_filename = osp.join(self.rundir, self.sh_entrypoint_filename)
        if osp.isfile(self.sh_entrypoint_filename):
            raise OSError("{} exists".format(self.sh_entrypoint_filename))
        sh = self.parse_sh_entrypoint()
        self.transfer_files["sh_entrypoint"] = self.sh_entrypoint_filename
        logger.info(
            "Dumping .sh entrypoint for cluster %s to %s",
            self.i_cluster,
            self.sh_entrypoint_filename,
        )
        if not (qondor.DRYMODE):
            with open(self.sh_entrypoint_filename, "w") as f:
                f.write(sh)

    def scope_to_file(self):
        self.scope_filename = osp.join(self.rundir, self.scope_filename)
        if osp.isfile(self.scope_filename):
            raise OSError("{} exists".format(self.scope_filename))
        self.transfer_files["scope"] = self.scope_filename
        self.env["QONDORSCOPEFILE"] = osp.basename(self.scope_filename)
        # Some last-minute additions before sending to a file
        self.scope["transfer_files"] = self.transfer_files
        self.scope["pips"] = self.pips
        logger.info(
            "Dumping the following scope for cluster %s to %s:\n%s",
            self.i_cluster,
            self.scope_filename,
            pprint.pformat(self.scope),
        )
        if not (qondor.DRYMODE):
            with open(self.scope_filename, "w") as f:
                json.dump(self.scope, f)

    def parse_sh_entrypoint(self):
        # Basic setup: Divert almost all output to the stderr, and setup cms scripts
        sh = [
            "#!/bin/bash",
            "set -e",
            'echo "hostname: $(hostname)"',
            'echo "date:     $(date)"',
            'echo "pwd:      $(pwd)"',
            'echo "ls -al:"',
            "ls -al",
            'echo "Redirecting all output to stderr from here on out"',
            "exec 1>&2",
            "",
            "export VO_CMS_SW_DIR=/cvmfs/cms.cern.ch/",
            "source /cvmfs/cms.cern.ch/cmsset_default.sh",
            "env > bare_env.txt",  # Save the environment before doing any other environment setup
            "",
        ]
        # Set the runtime environment (typically sourcing scripts to get the right python/gcc/ROOT/etc.)
        sh += self.run_env + [""]
        # Set up a directory to install python packages in, and put on the path
        # Currently requires $pipdir to be defined... might want to figure out something more clever
        sh += [
            "set -uxoE pipefail",
            'echo "Setting up custom pip install dir"',
            'HOME="$(pwd)"',
            'export pip_install_dir="$(pwd)/install"',
            'mkdir -p "${pip_install_dir}/bin"',
            'mkdir -p "${pip_install_dir}/lib/python2.7/site-packages"',
            'export PATH="${pip_install_dir}/bin:${PATH}"',
            "export PYTHONVERSION=$(python -c \"import sys; print('{}.{}'.format(sys.version_info.major, sys.version_info.minor))\")",
            'export PYTHONPATH="${pip_install_dir}/lib/python${PYTHONVERSION}/site-packages:${PYTHONPATH}"',
            "",
            "pip -V",
            "which pip",
            "",
        ]
        # `pip install` the required pip packages for the job
        for package, install_instruction in self.pips:
            package_name, version_stuff = qondor.utils.pip_split_version(
                package.rstrip("/")
            )
            package_name = package_name.replace(".", "-")
            package = package_name + version_stuff
            if version_stuff:
                install_instruction = (
                    "pypi"  # Force download from pypi for a specific version
                )
            if install_instruction == "auto":
                install_instruction = (
                    "editable" if qondor.utils.dist_is_editable(package) else "pypi"
                )
            if install_instruction == "editable":
                # Editable install: Manually give tarball, extract, and install
                sh.extend(
                    [
                        "mkdir {0}".format(package),
                        "tar xf {0}.tar -C {0}".format(package),
                        'pip install --install-option="--prefix=${{pip_install_dir}}" --no-cache-dir -e {0}/'.format(
                            package
                        ),
                    ]
                )
            else:
                # Non-editable install from pypi
                sh.append(
                    'pip install --install-option="--prefix=${{pip_install_dir}}" --no-cache-dir {0}'.format(
                        package
                    )
                )
        # Make the actual python call to run the required job code
        # Also echo the exitcode of the python command to a file, to easily check whether jobs succeeded
        # First compile the command - which might take some command line arguments
        python_cmd = "python {0}".format(osp.basename(self.runcode_filename))
        if self.run_args:
            # Add any arguments for the python script to this line
            try:  # py3
                from shlex import quote
            except ImportError:  # py2
                from pipes import quote
            python_cmd += " " + " ".join([quote(s) for s in self.run_args])
        sh += [
            python_cmd,
            'echo "$?" > exitcode_${QONDORCLUSTERNAME}_${CONDOR_CLUSTER_NUMBER}_${CONDOR_PROCESS_ID}.txt',  # Store the python exit code in a file
            "",
        ]
        sh = "\n".join(sh)
        logger.info(
            "Parsed the following .sh entrypoint for cluster %s:\n%s",
            self.i_cluster,
            sh,
        )
        return sh

    def submit(self, *args, **kwargs):
        """
        Like Session.submit(cluster, *args, **kwargs), but then initiated from the cluster object.
        If session is not provided as a keyword argument, a clean session is started.
        """
        session = kwargs.pop(
            "session", Session()
        )  # Make new session if not set by keyword
        session.submit(self, *args, **kwargs)


def cmsconnect_get_all_sites():
    """
    Reads the central config for cmsconnect to determine the list of all available sites
    """
    try:
        from configparser import RawConfigParser  # python 3
    except ImportError:
        import ConfigParser  # python 2

        RawConfigParser = ConfigParser.RawConfigParser
    cfg = RawConfigParser()
    cfg.read("/etc/ciconnect/config.ini")
    sites = cfg.get("submit", "DefaultSites").split(",")
    all_sites = set(sites)
    return all_sites


def cmsconnect_settings(sub, blacklist=None, whitelist=None, cli=False):
    """
    Adds special cmsconnect settings to submission dict in order to submit
    on cmsconnect via the python bindings, or in case of submitting by the
    command line, to set the DESIRED_Sites key.
    Modifies the dict in place.
    """
    all_sites = cmsconnect_get_all_sites()

    # Check whether the user whitelisted or blacklisted some sites
    desired_sites = None
    if blacklist or whitelist:
        import fnmatch

        blacklisted = []
        whitelisted = []
        # Build the blacklist
        if blacklist:
            for blacksite_pattern in blacklist:
                for site in all_sites:
                    if fnmatch.fnmatch(site, blacksite_pattern):
                        blacklisted.append(site)
        # Build the whitelist
        if whitelist:
            for whitesite_pattern in whitelist:
                for site in all_sites:
                    if fnmatch.fnmatch(site, whitesite_pattern):
                        whitelisted.append(site)
        # Convert to list and sort
        blacklisted = list(set(blacklisted))
        blacklisted.sort()
        whitelisted = list(set(whitelisted))
        whitelisted.sort()
        logger.info("Blacklisting: %s", ",".join(blacklisted))
        logger.info("Whitelisting: %s", ",".join(whitelisted))
        desired_sites = list(
            (set(all_sites) - set(blacklisted)).union(set(whitelisted))
        )
        desired_sites.sort()

    # Add a plus only if submitting via .jdl file
    def addplus(key):
        return "+" + key if cli else key

    if desired_sites:
        logger.info("Submitting to desired sites: %s", ",".join(desired_sites))
        sub[addplus("DESIRED_Sites")] = '"' + ",".join(desired_sites) + '"'
    else:
        logger.info("Submitting to all sites: %s", ",".join(all_sites))
    if not cli:
        sub[addplus("ConnectWrapper")] = '"2.0"'
        sub[addplus("CMSGroups")] = '"/cms,T3_US_FNALLPC"'
        sub[addplus("MaxWallTimeMins")] = "500"
        sub[addplus("ProjectName")] = '"cms.org.fnal"'
        sub[addplus("SubmitFile")] = '"irrelevant.jdl"'
        sub[addplus("AccountingGroup")] = '"analysis.{0}"'.format(os.environ["USER"])
        logger.warning(
            "FIXME: CMS Connect settings currently hard-coded for a FNAL user"
        )



"""
Module specific for the (boosted) svj analysis
"""

import logging
import os
import os.path as osp
import subprocess

import seutils

import qondor

logger = logging.getLogger("qondor")

MG_TARBALL_PATH = (
    "root://cmseos.fnal.gov//store/user/lpcsusyhad/SVJ2017/boosted/mg_tarballs"
)


class Physics(dict):
    """Holds the desired physics"""

    def __init__(self, *args, **kwargs):
        # Set some default values
        self["mz"] = 150.0
        self["rinv"] = 0.3
        self["boost"] = 0.0
        self["mdark"] = 20.0
        self["alpha"] = "peak"
        self["max_events"] = None
        super(Physics, self).__init__(*args, **kwargs)

    def boost_str(self):
        """Format string if boost > 0"""
        return "_HT{0:.0f}".format(self["boost"]) if self["boost"] > 0.0 else ""

    def max_events_str(self):
        return (
            ""
            if self.get("max_events", None) is None
            else "_n-{0}".format(self["max_events"])
        )


def svj_filename(step, physics):
    """
    Returns a basename for a root file that is input or output of a given step for given physics
    """
    rootfile = (
        "{step}_s-channel_mMed-{mz:.0f}_mDark-{mdark:.0f}_rinv-{rinv}_"
        "alpha-{alpha}{boost_str}_13TeV-madgraphMLM-pythia8{max_events_str}.root".format(
            step=step,
            boost_str=physics.boost_str(),
            max_events_str=physics.max_events_str(),
            **physics
        )
    )
    if physics.get("part", None):
        rootfile = rootfile.replace(".root", "_part-{}.root".format(physics["part"]))
    return rootfile


def madgraph_tarball_filename(physics):
    """Returns the basename of a MadGraph tarball for the given physics"""
    return svj_filename("step0_GRIDPACK", Physics(physics, part=None)).replace(
        ".root", ".tar.xz"
    )


def download_madgraph_tarball(physics, dst=None):
    """Downloads tarball from the storage element"""
    dst = osp.join(
        os.getcwd() if dst is None else dst, madgraph_tarball_filename(physics)
    )
    # Tarballs on SE will not have the boost tag and have postfix "_n-1"
    src = osp.join(
        MG_TARBALL_PATH,
        madgraph_tarball_filename(Physics(physics, boost=0.0, max_events=1)),
    )
    if osp.isfile(dst):
        logger.info("File %s already exists", dst)
    else:
        logger.info("Downloading %s --> %s", src, dst)
        seutils.cp(src, dst)


def step_cmd(inpre, outpre, physics):
    cmd = (
        "cmsRun runSVJ.py"
        " year={year}"
        " madgraph=1"
        " channel=s"
        " outpre={outpre}"
        " config={outpre}"
        " part={part}"
        " mMediator={mz:.0f}"
        " mDark={mdark:.0f}"
        " rinv={rinv}"
        " inpre={inpre}".format(inpre=inpre, outpre=outpre, **physics)
    )
    if "mingenjetpt" in physics:
        cmd += " mingenjetpt={0:.1f}".format(physics["mingenjetpt"])
    if "boost" in physics:
        cmd += " boost={0:.0f}".format(physics["boost"])
    if "max_events" in physics:
        cmd += " maxEvents={0}".format(physics["max_events"])
    return cmd


def gridpack_cmd(physics, nogridpack=False):
    cmd = (
        "python runMG.py"
        " year={year}"
        " madgraph=1"
        " channel=s"
        " outpre=step0_GRIDPACK"
        " mMediator={mz:.0f}"
        " mDark={mdark:.0f}"
        " rinv={rinv}".format(**physics)
    )
    if physics["boost"] > 0.0:
        cmd += " boost={}".format(physics["boost"])
    if physics["max_events"] > 0.0:
        cmd += " maxEvents={}".format(physics["max_events"])
    if nogridpack:
        cmd += " nogridpack=1"
    return cmd


class CMSSW(qondor.cmssw.CMSSW):
    """Subclass of the main CMSSW class for SVJ"""

    def __init__(self, *args, **kwargs):
        super(CMSSW, self).__init__(*args, **kwargs)
        self.svj_path = osp.join(self.cmssw_src, "SVJ/Production/test")

    def download_madgraph_tarball(self, physics):
        download_madgraph_tarball(physics, dst=self.svj_path)

    def _run_step(self, inpre, outpre, physics):
        """Runs the runSVJ script for 1 step"""
        expected_infile = osp.join(
            self.svj_path,
            madgraph_tarball_filename(physics)
            if inpre.startswith("step0")
            else svj_filename(inpre, physics),
        )
        expected_outfile = osp.join(self.svj_path, svj_filename(outpre, physics))
        if not osp.isfile(expected_infile):
            raise RuntimeError(
                "Expected input file {0} should exist now for step {1} -> {2}".format(
                    expected_infile, inpre, outpre
                )
            )
        self.run_commands(
            ["cd {0}".format(self.svj_path), step_cmd(inpre, outpre, physics)]
        )
        return expected_outfile

    def run_step(self, inpre, outpre, physics, n_attempts=1):
        """Wrapper around self._run_step with an n_attempts option"""
        i_attempt = 1
        while True:
            try:
                logger.info(
                    "Doing step %s -> %s (attempt %s/%s)",
                    inpre,
                    outpre,
                    i_attempt,
                    n_attempts,
                )
                expected_outfile = self._run_step(inpre, outpre, physics)
                return expected_outfile
            except subprocess.CalledProcessError:
                logger.error(
                    "Caught exception on step {0} -> {1}".format(inpre, outpre)
                )
                if i_attempt == n_attempts:
                    logger.error(
                        "step {0} -> {1} permanently failed".format(inpre, outpre)
                    )
                    raise
                else:
                    logger.error(
                        "This was attempt %s; Sleeping 60s and retrying", i_attempt
                    )
                    from time import sleep

                    sleep(60)
                    i_attempt += 1

    def run_chain(self, chain, physics, rootfile=None, move=False):
        """
        Runs a chain of steps. The first step is considered the input step (it is not ran).
        If rootfile is specified, it is copied into the svj_path. If move=True, it is moved instead (only
        possible for local filenames).
        """
        inpres = chain[:-1]
        outpres = chain[1:]
        # Copy/move the input rootfile if it's given
        if rootfile:
            expected_infile = osp.join(self.svj_path, svj_filename(inpres[0], physics))
            if move:
                logger.info("Moving %s -> %s", rootfile, expected_infile)
                os.rename(rootfile, expected_infile)
            else:
                seutils.cp(rootfile, expected_infile)
        # Run steps
        for inpre, outpre in zip(inpres, outpres):
            expected_outfile = self.run_step(
                inpre,
                outpre,
                physics,
                n_attempts=3 if ("RECO" in outpre or "DIGI" in outpre) else 1,
            )
        return expected_outfile

    def run_gridpack(self, physics, nogridpack=False):
        """
        Creates a MadGraph gridpack for the given physics
        """
        return self.run_commands(
            [
                "cd {0}".format(self.svj_path),
                gridpack_cmd(physics, nogridpack=nogridpack),
            ]
        )


class TreeMakerCMSSW(qondor.cmssw.CMSSW):
    """Subclass of the main CMSSW class for SVJ"""

    _readfiles_cache = {}

    def __init__(self, *args, **kwargs):
        super(TreeMakerCMSSW, self).__init__(*args, **kwargs)
        self.treemaker_path = osp.join(self.cmssw_src, "TreeMaker/Production/test")

    @classmethod
    def get_readfiles(cls, bkg):
        """
        Hacky: gets the readFiles from the main TreeMaker repo.
        Caches results in class variable so subsequent calls just use the cache
        """
        if bkg in cls._readfiles_cache:
            return cls._readfiles_cache[bkg]
        import re

        scenario, bkg_string = bkg.split(".", 1)
        url = "https://raw.githubusercontent.com/TreeMaker/TreeMaker/Run2_2017/Production/python/{}/{}_cff.py".format(
            scenario, bkg_string
        )
        text = qondor.utils.strip_comments(qondor.utils.download_url_to_str(url))
        rootfiles = re.findall(r"/store/mc.*?root", text)
        cls._readfiles_cache[
            bkg
        ] = rootfiles  # Cache result so we can safely call the method again
        return rootfiles

    @classmethod
    def count_readfiles(cls, bkg):
        return len(cls.get_readfiles(bkg))

    def command_bkg(self, bkg, i_file, n_events=None):
        scenario = bkg.split(".")[0]
        cmd = (
            "cmsRun runMakeTreeFromMiniAOD_cfg.py"
            " outfile=outfile"
            " scenario={}"
            " inputFilesConfig={}"
            " lostlepton=0"
            " doZinv=0"
            " systematics=0"
            " deepAK8=0"
            " deepDoubleB=0"
            " doPDFs=0"
            " nestedVectors=False"
            " splitLevel=99"
            " nstart={}"
            " nfiles=1".format(scenario, bkg, i_file)
        )
        if n_events:
            cmd += " numevents={}".format(n_events)
        return cmd

    def run_bkg(self, *args, **kwargs):
        self.run_commands(
            ["cd {0}".format(self.treemaker_path), self.command_bkg(*args, **kwargs)]
        )
        expected_outfile = osp.join(self.treemaker_path, "outfile_RA2AnalysisTree.root")
        return expected_outfile


def init_cmssw(
    tarball_key="cmssw_tarball", scram_arch=None, outdir=None, init_class=CMSSW
):
    """
    Like the main qondor.init_cmssw, but for qondor.svj.CMSSW
    """
    if osp.isfile(tarball_key):
        # A path to a local tarball was given
        cmssw_tarball = tarball_key
    elif seutils.has_protocol(tarball_key):
        # A path to a tarball on a storage element was given
        cmssw_tarball = tarball_key
    else:
        # A key to a file in the preprocessing was given
        cmssw_tarball = qondor.get_preproc().files[tarball_key]
    cmssw = init_class.from_tarball(cmssw_tarball, scram_arch, outdir=outdir)
    return cmssw


def init_cmssw_treemaker(*args, **kwargs):
    return init_cmssw(init_class=TreeMakerCMSSW, *args, **kwargs)



# -*- coding: utf-8 -*-
import argparse
import json
import logging
import os
import os.path as osp
import pprint

from .logger import colored, setup_logger, setup_subprocess_logger

logger = setup_logger()
subprocess_logger = setup_subprocess_logger()


def debug(flag=True):
    if flag:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


# Path to non-python-package files
INCLUDE_DIR = osp.join(osp.abspath(osp.dirname(__file__)), "include")

# Load version as a module variable
with open(osp.join(INCLUDE_DIR, "VERSION"), "r") as f:
    version = f.read().strip()

# Check if qondor is imported in batch mode
BATCHMODE = False
if "QONDOR_BATCHMODE" in os.environ:
    BATCHMODE = True

# Global variable to check if the htcondor bindings are installed
try:
    import htcondor

    logger.debug("The python bindings for htcondor are imported")
    BINDINGS_INSTALLED = True
except ImportError:
    logger.debug("The python bindings for htcondor do not seem to be installed")
    BINDINGS_INSTALLED = False

COLLECTOR_NODES = None
DEFAULT_MGM = None
TIMESTAMP_FMT = "%Y%m%d_%H%M%S"

import seutils

from . import cmssw, resubmit, schedd, svj, utils
from .cmssw import CMSSW
from .cmssw_releases import get_arch
from .submit import get_first_cluster

# ___________________________________________________
# Decisions at import-time depending on whether this is a job or
# an import on an interactive node


def _json_load_byteified(file_handle):
    """
    object_hook method to load regular strings rather than unicode strings
    when doing json.load(s)
    See https://stackoverflow.com/a/33571117/9209944
    """
    return _byteify(json.load(file_handle, object_hook=_byteify), ignore_dicts=True)


def _json_loads_byteified(json_text):
    return _byteify(json.loads(json_text, object_hook=_byteify), ignore_dicts=True)


def _byteify(data, ignore_dicts=False):
    try:
        # if this is a unicode string, return its string representation
        # Only for python 2, NameError in python 3
        if isinstance(data, unicode):
            return data.encode("utf-8")
    except NameError:
        pass
    # if this is a list of values, return list of byteified values
    if isinstance(data, list):
        return [_byteify(item, ignore_dicts=True) for item in data]
    # if this is a dictionary, return dictionary of byteified keys and values
    # but only if we haven't already byteified it
    if isinstance(data, dict) and not ignore_dicts:
        # Python 2.6 compatibility, no dict comprehensions
        return dict(
            [
                (_byteify(key, ignore_dicts=True), _byteify(value, ignore_dicts=True))
                for key, value in data.items()
            ]
        )
    # if it's anything else, return it in its original form
    return data


_TRIED_ONCE = False


class Scope(argparse.Namespace):
    """
    Like an argparse.Namespace object, but when first called loads the scope either
    from a .json file or from the python file on top of the stack (first cluster).
    """

    def __init__(self, *args, **kwargs):
        super(Scope, self).__init__(*args, **kwargs)
        self.is_loaded = False

    def load_batchmode(self):
        if "QONDORSCOPEFILE" in os.environ:
            scope_file = os.environ["QONDORSCOPEFILE"]
            if osp.isfile(scope_file):
                with open(scope_file, "r") as f:
                    # d = json.load(f)
                    d = _json_load_byteified(f)
                    scope.__dict__.update(d)
                self.is_loaded = True
                logger.info(
                    "Loaded following scope from %s:\n%s",
                    scope_file,
                    pprint.pformat(scope),
                )
                return
        logger.info("Could not load scope")

    def load_localmode(self):
        import traceback

        stack = traceback.extract_stack()
        python_file = stack[0][0]
        cluster = get_first_cluster(python_file)
        self.__dict__.update(cluster.scope)
        self.is_loaded = True

    def load(self):
        logger.info("Trying to load scope...")
        self.load_batchmode() if BATCHMODE else self.load_localmode()

    # argparse.Namespace doesn't have a __getitem__ method
    # def __getitem__(self, *args, **kwargs):
    #     """
    #     Hacked to call .load() once upon the first call of __getitem__
    #     """
    #     global _TRIED_ONCE
    #     if not _TRIED_ONCE:
    #         _TRIED_ONCE = True
    #         self.load()
    #     return super(Scope, self).__getitem__(*args, **kwargs)

    def __getattribute__(self, *args, **kwargs):
        """
        Hacked to call .load() once upon the first call of __getattribute__
        """
        global _TRIED_ONCE
        if not _TRIED_ONCE:
            _TRIED_ONCE = True
            self.load()
        return super(Scope, self).__getattribute__(*args, **kwargs)


scope = Scope()


def load_seutils_cache():
    if BATCHMODE and scope.is_loaded and "seutils-cache" in scope.transfer_files:
        seutils_cache_tarball = osp.basename(scope.transfer_files["seutils-cache"])
        if osp.isfile(seutils_cache_tarball):
            seutils.load_tarball_cache(seutils_cache_tarball)


if BATCHMODE:
    load_seutils_cache()
    seutils.N_COPY_RETRIES = (
        2  # Sets default amount of times to retry a seutils.cp statement
    )

# Read rootcache is this is a job
if BATCHMODE and osp.isfile("rootcache.tar.gz"):
    seutils.root.load_cache("rootcache.tar.gz")

# FNAL-specific things
if os.environ.get("HOSTNAME", "").endswith("fnal.gov"):
    # Fix to be able to import htcondor python bindings
    import sys

    try:
        major_version = sys.version_info.major
    except AttributeError:  # Python 2.6 compatibility
        major_version = sys.version_info[1]
    if major_version < 3:
        logger.warning(
            "Detected FNAL: Modifying path to use system htcondor python bindings"
        )
        sys.path.extend(
            ["/usr/lib64/python2.6/site-packages", "/usr/lib64/python2.7/site-packages"]
        )
    schedd.GLOBAL_SCHEDDMAN_CLS = schedd.ScheddManagerFermiHTC
    seutils.set_default_mgm("root://cmseos.fnal.gov")

# Load the bare environment saved before doing any setup
BARE_ENV = {}
if BATCHMODE and osp.isfile("bare_env.txt"):
    with open("bare_env.txt") as f:
        for line in f.readlines():
            key, value = line.strip().split("=", 1)
            BARE_ENV[key] = value

# ___________________________________________________
# 'Globals'

DRYMODE = False


def drymode(flag=True):
    global DRYMODE
    DRYMODE = flag


# ___________________________________________________
# Short cuts


def get_proc_id():
    if BATCHMODE:
        if "QONDOR_PROC_ID_RESUBMISSION" in os.environ:
            # This is a resubmission. Return the i_proc for the resubmission instead
            i_proc = int(os.environ["QONDOR_PROC_ID_RESUBMISSION"])
            logger.warning("This is a resubmission - returning i_proc = %s", i_proc)
            return i_proc
        return int(os.environ["CONDOR_PROCESS_ID"])
    else:
        logger.info("Local mode - return proc_id 0")
        return 0


def get_cluster_id():
    if BATCHMODE:
        return os.environ["CONDOR_CLUSTER_NUMBER"]
    else:
        logger.info("Local mode - return cluster_id 1234")
        return str(1234)


def init_cmssw(tarball_key="cmssw_tarball", scram_arch=None, outdir=None):
    """
    A shortcut function to quickly extract and setup a CMSSW tarball.
    The first argument `tarball_key` may also be a path (either on a storage element or local).
    """
    if osp.isfile(tarball_key):
        # A path to a local tarball was given
        cmssw_tarball = tarball_key
    elif seutils.has_protocol(tarball_key):
        # A path to a tarball on a storage element was given
        cmssw_tarball = tarball_key
    elif tarball_key in scope:
        # A key to a file in the preprocessing was given
        cmssw_tarball = scope[tarball_key]
    else:
        raise Exception("Cannot load tarball {0}".format(tarball_key))
    return CMSSW.from_tarball(cmssw_tarball, scram_arch, outdir=outdir)


def get_submission_timestamp():
    """
    Returns the submission time as a timestamp string with format TIMESTAMP_FMT
    """
    if BATCHMODE:
        return os.environ["CLUSTER_SUBMISSION_TIMESTAMP"]
    else:
        from datetime import datetime

        logger.info("Local mode - returning current time")
        return datetime.now().strftime(TIMESTAMP_FMT)


def get_submission_time():
    """
    Returns the submission time as datetime object
    """
    from datetime import datetime

    if BATCHMODE:
        return datetime.strptime(get_submission_timestamp(), TIMESTAMP_FMT)
    else:
        logger.info("Local mode - returning current time")
        return datetime.now()


def get_submission_timestr(fmt="%b%d"):
    return get_submission_time().strftime(fmt)
